@@ -36,6 +36,8 @@ class BoardInterface:
 		            'jlink_device': 'nrf52',
 		            'page_size': 4096,
 		            'openocd': 'nordic_nrf52_dk.cfg'},
+		'nano33ble': {'description': 'Arduino Nano 33 BLE board',
+		              'arch': 'cortex-m4'},
 		'launchxl-cc26x2r1': {'description': 'TI CC26x2-based launchpad',
 		                      'arch': 'cortex-m4',
 		                      'page_size': 512,
@@ -77,6 +79,33 @@ class BoardInterface:
 		         'openocd_commands': {'program': 'jtagspi_program {{binary}} {address:#x};',
 		                              'read': 'jtagspi_read {{binary}} {address:#x} {length};',
 		                              'erase': 'flash fillb {address:#x} 0x00 512;'}},
+		'stm32f3discovery': {'description': 'STM32F3-based Discovery Boards',
+		                     'arch': 'cortex-m4',
+		                     'apps_start_address': 0x08020000,
+		                     'page_size': 2048,
+		                     'openocd': 'external',
+		                     'openocd_prefix': 'interface hla; \
+		                                        hla_layout stlink; \
+		                                        hla_device_desc "ST-LINK/V2-1"; \
+		                                        hla_vid_pid 0x0483 0x374b; \
+		                                        set WORKAREASIZE 0xC000; \
+		                                        source [find target/stm32f3x.cfg];'},
+		'stm32f4discovery': {'description': 'STM32F4-based Discovery Boards',
+		                        'arch': 'cortex-m4',
+		                        'apps_start_address': 0x08040000,
+		                        'page_size': 2048,
+		                        'openocd': 'external',
+		                        'openocd_prefix': 'interface hla; \
+		                                           hla_layout stlink; \
+		                                           hla_device_desc "ST-LINK/V2-1"; \
+		                                           hla_vid_pid 0x0483 0x374b; \
+		                                           set WORKAREASIZE 0x40000; \
+		                                           source [find target/stm32f4x.cfg];'},
+		'nucleof4': {'description': 'STM32f4-based Nucleo development boards',
+		             'arch': 'cortex-m4',
+		             'apps_start_address': 0x08040000,
+		             'page_size': 2048,
+		             'openocd': 'st_nucleo_f4.cfg'},
 		'hifive1': {'description': 'SiFive HiFive1 development board',
 		            'arch': 'rv32imac',
 		            'apps_start_address': 0x20430000,
@@ -87,7 +116,8 @@ class BoardInterface:
 		             'apps_start_address': 0x20040000,
 		             'page_size': 512,
 		             'jlink_device': 'FE310',
-		             'jlink_if': 'jtag'},
+		             'jlink_if': 'jtag',
+		             'openocd': 'sifive-hifive1-revb.cfg'},
 		'edu-ciaa': {'description': 'Educational NXP board, from the CIAA project',
 		             'arch': 'cortex-m4',
 		             'page_size': 512,
@@ -96,6 +126,17 @@ class BoardInterface:
 		             'openocd_options': ['noreset'],
 		             'openocd_commands': {'program': 'flash write_image erase {{binary}} {address:#x};verify_image {{binary}} {address:#x};',
 		             'erase': 'flash fillb {address:#x} 0x00 512;'}},
+		'microbit_v2': {'description': 'BBC Micro:bit v2',
+		                'arch': 'cortex-m4',
+		                'apps_start_address': 0x00040000,
+		                'page_size': 4096,
+		                'openocd': 'external',
+		                'openocd_prefix': 'source [find interface/cmsis-dap.cfg]; \
+		                                   transport select swd; \
+		                                   source [find target/nrf52.cfg]; \
+		                                   set WORKAREASIZE 0x40000; \
+		                                   $_TARGETNAME configure -work-area-phys 0x20000000 -work-area-size $WORKAREASIZE -work-area-backup 0; \
+		                                   flash bank $_CHIPNAME.flash nrf51 0x00000000 0 1 1 $_TARGETNAME;'},
 	}
 
 	def __init__ (self, args):
@@ -152,6 +193,14 @@ class BoardInterface:
 		# methods need. There may be flags specific to a particular
 		# communication interface.
 
+	def attached_board_exists (self):
+		'''
+		For this particular board communication channel, check if there appears
+		to be a valid board attached to the host that tockloader can communicate
+		with.
+		'''
+		return False
+
 	def open_link_to_board (self):
 		'''
 		Open a connection to the board.
@@ -182,8 +231,7 @@ class BoardInterface:
 
 		If this fails for some reason this should return an empty binary array.
 		'''
-		if self.args.debug:
-			logging.debug('DEBUG => Read Range, address: {:#010x}, length: {}'.format(address, length))
+		logging.debug('DEBUG => Read Range, address: {:#010x}, length: {}'.format(address, length))
 
 	def erase_page (self, address):
 		'''
@@ -205,21 +253,41 @@ class BoardInterface:
 		'''
 		Get all attributes on a board. Returns an array of attribute dicts.
 		'''
+		# Check for cached attributes.
+		if hasattr(self, 'attributes'):
+			return self.attributes
+
 		# Read the entire block of attributes directly from flash.
 		# This is much faster.
 		def chunks(l, n):
 			for i in range(0, len(l), n):
 				yield l[i:i + n]
 		raw = self.read_range(0x600, 64*16)
-		return [self._decode_attribute(r) for r in chunks(raw, 64)]
+		attributes = [self._decode_attribute(r) for r in chunks(raw, 64)]
+
+		# Cache what we get in case this gets called again.
+		self.attributes = attributes
+
+		return attributes
 
 	def set_attribute (self, index, raw):
 		'''
 		Set a single attribute.
 		'''
+		# Remove any cached attributes
+		del self.attributes
+
 		address = 0x600 + (64 * index)
 		self.flash_binary(address, raw)
 
+	def set_start_address (self, address):
+		'''
+		Set the address the bootloader jumps to to start the actual code.
+		'''
+		# This is only valid if there is a bootloader and this function is
+		# re-implemented.
+		raise TockLoaderException('No bootloader, cannot set start address.')
+
 	def _decode_attribute (self, raw):
 		try:
 			key = raw[0:8].decode('utf-8').strip(bytes([0]).decode('utf-8'))
@@ -276,6 +344,23 @@ class BoardInterface:
 		self.apps_start_address = 0x30000
 		return self.apps_start_address
 
+	def get_kernel_version (self):
+		'''
+		Return the kernel ABI version installed on the board. If the version
+		cannot be determined, return `None`.
+		'''
+		# Check if there is an attribute we can use.
+		attributes = self.get_all_attributes()
+		for attribute in attributes:
+			if attribute and attribute['key'] == 'kernver':
+				kernver = attribute['value'].strip()
+				logging.debug('Determined kernel version is "{}".'.format(kernver))
+				return kernver
+
+		# If not in an attribute we give up and return None.
+		logging.debug('Could not determine kernel version.')
+		return None
+
 	def determine_current_board (self):
 		'''
 		Figure out which board we are connected to. Most likely done by reading
