@@ -214,6 +214,14 @@ def command_flash (args):
 			binary += f.read()
 		count += 1
 
+	# Check if the user asked us to pad the binary with some additional bytes.
+	pad = None
+	if args.pad:
+		# First arg is the length, second arg is the value
+		pad = (args.pad[0], args.pad[1])
+		if pad[1] < 0 or pad[1] > 255:
+			raise TockLoaderException('Padding value must be only one byte')
+
 	# Flash the binary to the chip
 	tock_loader = TockLoader(args)
 	tock_loader.open()
@@ -222,7 +230,7 @@ def command_flash (args):
 	if count > 1:
 		plural = 'ies'
 	logging.status('Flashing binar{} to board...'.format(plural))
-	tock_loader.flash_binary(binary, args.address)
+	tock_loader.flash_binary(binary, args.address, pad=pad)
 
 
 def command_read (args):
@@ -238,6 +246,24 @@ def command_read (args):
 	tock_loader.read_flash(args.address, args.length)
 
 
+def command_write (args):
+	'''
+	Write flash range on the chip with a specific value.
+	'''
+	tock_loader = TockLoader(args)
+	tock_loader.open()
+
+	# Only write a single byte.
+	if args.value < 0 or args.value > 0xff:
+		raise TockLoaderException('Can only write multiple copies of a single byte')
+
+	logging.status('Writing flash on the board...')
+	logging.status('  Address: {:#x}'.format(args.address))
+	logging.status('  Length:  {} bytes'.format(args.length))
+	logging.status('  Value:   {:#x}'.format(args.value))
+	tock_loader.write_flash(args.address, args.length, args.value)
+
+
 def command_list_attributes (args):
 	tock_loader = TockLoader(args)
 	tock_loader.open()
@@ -262,6 +288,14 @@ def command_remove_attribute (args):
 	tock_loader.remove_attribute(args.key)
 
 
+def command_set_start_address (args):
+	tock_loader = TockLoader(args)
+	tock_loader.open()
+
+	logging.status('Setting bootloader jump address...')
+	tock_loader.set_start_address(args.address)
+
+
 def command_info (args):
 	tock_loader = TockLoader(args)
 	tock_loader.open()
@@ -328,7 +362,7 @@ def main ():
 	atexit.register(helpers.set_terminal_title, '')
 
 	# Setup logging for displaying background information to the user.
-	logging.basicConfig(style='{', format='[{levelname:<7}] {message}', level=logging.DEBUG)
+	logging.basicConfig(style='{', format='[{levelname:<7}] {message}', level=logging.INFO)
 	# Add a custom status level for logging what tockloader is doing.
 	logging.addLevelName(25, 'STATUS')
 	logging.Logger.status = functools.partialmethod(logging.Logger.log, 25)
@@ -372,6 +406,9 @@ def main ():
 	parent_channel.add_argument('--port', '-p', '--device', '-d',
 		help='The serial port or device name to use',
 		metavar='STR')
+	parent_channel.add_argument('--serial',
+		action='store_true',
+		help='Use the serial bootloader to flash.')
 	parent_channel.add_argument('--jtag',
 		action='store_true',
 		help='Use JTAG and JLinkExe to flash. Deprecated. Use --jlink instead.')
@@ -431,15 +468,17 @@ def main ():
 		title='Commands',
 		metavar='')
 
+	# Command Groups
+	#
+	# Python argparse doesn't support grouping commands in subparsers as of
+	# January 2021 :(. The best we can do now is order them logically.
+
 	listen = subparser.add_parser('listen',
 		parents=[parent],
 		help='Open a terminal to receive UART data')
 	listen.add_argument('--port', '-p', '--device', '-d',
 		help='The serial port or device name to use',
 		metavar='STR')
-	listen.add_argument('--wait-to-listen',
-		help='Wait until contacted on server socket to actually listen',
-		action='store_true')
 	listen.add_argument('--timestamp',
 		help='Prepend output with a timestamp',
 		action='store_true')
@@ -467,17 +506,6 @@ def main ():
 		help='The interface type to pass to JLinkExe. Only used with --rtt.')
 	listen.set_defaults(func=command_listen)
 
-	listcmd = subparser.add_parser('list',
-		parents=[parent, parent_apps, parent_channel],
-		help='List the apps installed on the board')
-	listcmd.set_defaults(func=command_list)
-	listcmd.add_argument('--verbose', '-v',
-		help='Print more information',
-		action='store_true')
-	listcmd.add_argument('--quiet', '-q',
-		help='Print just a list of application names',
-		action='store_true')
-
 	install = subparser.add_parser('install',
 		parents=[parent, parent_apps, parent_channel],
 		help='Install apps on the board')
@@ -514,6 +542,22 @@ def main ():
 		help='The name of the app(s) to remove',
 		nargs='*')
 
+	listcmd = subparser.add_parser('list',
+		parents=[parent, parent_apps, parent_channel],
+		help='List the apps installed on the board')
+	listcmd.set_defaults(func=command_list)
+	listcmd.add_argument('--verbose', '-v',
+		help='Print more information',
+		action='store_true')
+	listcmd.add_argument('--quiet', '-q',
+		help='Print just a list of application names',
+		action='store_true')
+
+	info = subparser.add_parser('info',
+		parents=[parent, parent_apps, parent_channel],
+		help='Verbose information about the connected board')
+	info.set_defaults(func=command_info)
+
 	eraseapps = subparser.add_parser('erase-apps',
 		parents=[parent, parent_apps, parent_channel],
 		help='Delete apps from the board')
@@ -562,20 +606,53 @@ def main ():
 		help='Address to flash the binary at',
 		type=lambda x: int(x, 0),
 		default=0x30000)
+	flash.add_argument('--pad',
+		help='Optional number of bytes to pad and value to use',
+		nargs=2,
+		type=lambda x: int(x, 0),)
+	flash.add_argument('--set-attribute',
+		help='Key-value attribute pair to set if not already set',
+		nargs=2,
+		action='append')
 
-	flash = subparser.add_parser('read',
+	read = subparser.add_parser('read',
 		parents=[parent, parent_channel],
 		help='Read arbitrary flash memory')
-	flash.set_defaults(func=command_read)
-	flash.add_argument('address',
+	read.set_defaults(func=command_read)
+	read.add_argument('address',
 		help='Address to read from',
 		type=lambda x: int(x, 0),
 		default=0x30000)
-	flash.add_argument('length',
+	read.add_argument('length',
 		help='Number of bytes to read',
 		type=lambda x: int(x, 0),
 		default=512)
 
+	write = subparser.add_parser('write',
+		parents=[parent, parent_channel],
+		help='Write arbitrary flash memory with constant value')
+	write.set_defaults(func=command_write)
+	write.add_argument('address',
+		help='Address to write to',
+		type=lambda x: int(x, 0),
+		default=0x30000)
+	write.add_argument('length',
+		help='Number of bytes to write',
+		type=lambda x: int(x, 0),
+		default=512)
+	write.add_argument('value',
+		help='Value to write',
+		type=lambda x: int(x, 0),
+		default=0xff)
+
+	dump_flash_page = subparser.add_parser('dump-flash-page',
+		parents=[parent, parent_channel],
+		help='Read a page of flash from the board')
+	dump_flash_page.set_defaults(func=command_dump_flash_page)
+	dump_flash_page.add_argument('page',
+		help='The number of the page to read',
+		type=lambda x: int(x, 0))
+
 	listattributes = subparser.add_parser('list-attributes',
 		parents=[parent, parent_channel],
 		help='List attributes stored on the board')
@@ -597,10 +674,14 @@ def main ():
 	removeattribute.add_argument('key',
 		help='Attribute key')
 
-	info = subparser.add_parser('info',
-		parents=[parent, parent_apps, parent_channel],
-		help='Verbose information about the connected board')
-	info.set_defaults(func=command_info)
+	setstartaddress = subparser.add_parser('set-start-address',
+		parents=[parent, parent_channel],
+		help='Set bootloader jump address')
+	setstartaddress.set_defaults(func=command_set_start_address)
+	setstartaddress.add_argument('address',
+		help='Start address',
+		type=lambda x: int(x, 0),
+		default=0x10000)
 
 	inspect_tab = subparser.add_parser('inspect-tab',
 		parents=[parent],
@@ -613,20 +694,18 @@ def main ():
 		help='The TAB or TABs to inspect',
 		nargs='*')
 
-	dump_flash_page = subparser.add_parser('dump-flash-page',
-		parents=[parent, parent_channel],
-		help='Read a page of flash from the board')
-	dump_flash_page.set_defaults(func=command_dump_flash_page)
-	dump_flash_page.add_argument('page',
-		help='The number of the page to read',
-		type=lambda x: int(x, 0))
-
 	list_known_boards = subparser.add_parser('list-known-boards',
 		help='List the boards that Tockloader explicitly knows about')
 	list_known_boards.set_defaults(func=command_list_known_boards)
 
 	argcomplete.autocomplete(parser)
-	args = parser.parse_args()
+	args, unknown_args = parser.parse_known_args()
+
+	# Warn about unknown arguments, suggest tockloader update.
+	if len(unknown_args) > 0:
+		logging.warning('Unknown arguments passed. You may need to update tockloader.')
+		for unknown_arg in unknown_args:
+			logging.warning('Unknown argument "{}"'.format(unknown_arg))
 
 	# Concat the args before the command with those that were specified
 	# after the command. This is a workaround because for some reason python
@@ -636,6 +715,10 @@ def main ():
 		if getattr(args, key) != value:
 			setattr(args, key, value)
 
+	# Change logging level if `--debug` was supplied.
+	if args.debug:
+		logging.getLogger('').setLevel(logging.DEBUG)
+
 	# Handle deprecated arguments.
 	# --jtag is now --jlink. If --jtag was passed copy it to --jlink.
 	if hasattr(args, 'jtag') and args.jtag:
