@@ -87,8 +87,8 @@ class TAB:
 
 	def extract_tbf (self, tbf_name):
 		'''
-		Return a `TabApp` object from this TAB. You must specify the desired TBF
-		name, and only that TBF will be returned.
+		Return a `TabApp` object from this TAB. You must specify the
+		desired TBF name, and only that TBF will be returned.
 		'''
 		tbf_filename = '{}.tbf'.format(tbf_name)
 		binary_tarinfo = self.tab.getmember(tbf_filename)
@@ -113,21 +113,34 @@ class TAB:
 		'''
 		Check if the Tock app is compatible with a particular Tock board.
 		'''
-		metadata = self.parse_metadata()
-		if metadata['tab-version'] == 1:
-			return 'only-for-boards' not in metadata or \
-			       board in metadata['only-for-boards'] or \
-			       metadata['only-for-boards'] == ''
-		else:
-			raise TockLoaderException('Unable to understand version {} of metadata'.format(metadata['tab-version']))
+		only_for_boards = self._get_metadata_key('only-for-boards')
+		return only_for_boards == None or \
+		       board in only_for_boards or \
+		       only_for_boards == ''
 
-	def parse_metadata (self):
+	def get_compatible_boards (self):
 		'''
-		Open and parse the included metadata file in the TAB.
+		Return a list of compatible boards from the metadata file.
 		'''
-		metadata_tarinfo = self.tab.getmember('metadata.toml')
-		metadata_str = self.tab.extractfile(metadata_tarinfo).read().decode('utf-8')
-		return pytoml.loads(metadata_str)
+		only_for_boards = self._get_metadata_key('only-for-boards') or ''
+		return [b.strip() for b in only_for_boards.split(',')]
+
+	def is_compatible_with_kernel_version (self, kernel_version):
+		'''
+		Check if the Tock app is compatible with the version of the kernel.
+		Default to yes unless we can be confident the answer is no.
+
+		`kernel_version` should be a string, or None if the kernel API version
+		is unknown.
+		'''
+		if kernel_version == None:
+			# Bail out early if kernel version is unknown.
+			return True
+
+		tock_kernel_version = self._get_metadata_key('tock-kernel-version')
+		return tock_kernel_version == None or \
+		       tock_kernel_version == kernel_version or \
+		       tock_kernel_version == ''
 
 	def get_supported_architectures (self):
 		'''
@@ -163,9 +176,45 @@ class TAB:
 				tbfs.append(f[:-4])
 		return tbfs
 
+	def get_app_name (self):
+		'''
+		Return the app name from the metadata file.
+		'''
+		return self._get_metadata_key('name') or ''
+
+	def _parse_metadata (self):
+		'''
+		Open and parse the included metadata file in the TAB, returning the
+		key-value pairs as a dict.
+		'''
+		# Use cached value.
+		if hasattr(self, 'metadata'):
+			return self.metadata
+
+		# Otherwise parse f.toml file.
+		metadata_tarinfo = self.tab.getmember('metadata.toml')
+		metadata_str = self.tab.extractfile(metadata_tarinfo).read().decode('utf-8')
+		self.metadata = pytoml.loads(metadata_str)
+		return self.metadata
+
+	def _get_metadata_key (self, key):
+		'''
+		Return the value for a specific key from the metadata file.
+		'''
+		metadata = self._parse_metadata()
+		if metadata['tab-version'] == 1:
+			if key in metadata:
+				return metadata[key].strip()
+		else:
+			# If unknown version, print warning.
+			logging.warning('Unknown TAB version. You probably need to update tockloader.')
+
+		# Default to returning None if key is not found or file is not understood.
+		return None
+
 	def __str__ (self):
 		out = ''
-		metadata = self.parse_metadata()
+		metadata = self._parse_metadata()
 		out += 'TAB: {}\n'.format(metadata['name'])
 		for k,v in sorted(metadata.items()):
 			if k == 'name':
