@@ -8,6 +8,7 @@ channel specific code is in other files.
 import binascii
 import contextlib
 import copy
+import ctypes
 import functools
 import itertools
 import logging
@@ -71,7 +72,8 @@ class TockLoader:
 	    'nrf52dk': {'size_minimum': 4096},
 	    'edu-ciaa': {'cmd_flags': {'bundle_apps': True,
 	                               'openocd': True}},
-	    'arty': {'size_minimum': 0x10000},
+	    'arty': {'size_minimum': 0x10000,
+	             'size_constraint': None},
 	}
 
 
@@ -100,28 +102,87 @@ class TockLoader:
 
 		# Get an object that allows talking to the board.
 		if hasattr(self.args, 'jlink') and self.args.jlink:
+			# User passed `--jlink`. Force the jlink channel.
 			self.channel = JLinkExe(self.args)
 		elif hasattr(self.args, 'openocd') and self.args.openocd:
+			# User passed `--openocd`. Force the OpenOCD channel.
 			self.channel = OpenOCD(self.args)
-		else:
+		elif hasattr(self.args, 'serial') and self.args.serial:
+			# User passed `--serial`. Force the serial bootloader channel.
 			self.channel = BootloaderSerial(self.args)
+		else:
+			# Try to do some magic to determine the correct channel to use. Our
+			# goal is to automatically choose the correct setting so that
+			# `tockloader` just works without having to specify a board and any
+			# flags.
+
+			# Loop so we can `break`. This will never execute more than once.
+			while (True):
+
+				# One issue is that JTAG connections often expose both a JTAG
+				# and a serial port. So, if we try to use the serial port first
+				# we will incorrectly detect that serial port. So, we start with
+				# the less likely jtag channel. We start with jlinkexe because
+				# it has been in tockloader longer. Let me know if this is the
+				# wrong decision.
+				jlink_channel = JLinkExe(self.args)
+				if jlink_channel.attached_board_exists():
+					self.channel = jlink_channel
+					logging.info('Using jlink channel to communicate with the board.')
+					break
+
+				# Next try openocd.
+				openocd_channel = OpenOCD(self.args)
+				if openocd_channel.attached_board_exists():
+					self.channel = openocd_channel
+					logging.info('Using openocd channel to communicate with the board.')
+					break
+
+				# Default to using the serial bootloader. This is how tockloader
+				# has worked for a long time. We may want to change this, but
+				# for now we don't change the default behavior.
+				self.channel = BootloaderSerial(self.args)
+
+				# Exit while(1) loop, do not remove this break!
+				break
 
 		# And make sure the channel is open (e.g. open a serial port).
 		self.channel.open_link_to_board()
 
 
-	def flash_binary (self, binary, address):
+	def flash_binary (self, binary, address, pad=None):
 		'''
 		Tell the bootloader to save the binary blob to an address in internal
 		flash.
 
 		This will pad the binary as needed, so don't worry about the binary
 		being a certain length.
+
+		This accepts an optional `pad` parameter. If used, the `pad` parameter
+		is a tuple of `(length, value)` signifying the number of bytes to pad,
+		and the particular byte to use for the padding.
 		'''
 		# Enter bootloader mode to get things started
 		with self._start_communication_with_board():
+
+			# Check if we should add padding, which is just pad[0] copies of the
+			# same byte (pad[1]).
+			if pad:
+				extra = bytes([pad[1]] * pad[0])
+				binary = binary + extra
+
 			self.channel.flash_binary(address, binary)
 
+			# Flash command can optionally set attributes.
+			if self.args.set_attribute != None:
+				for k,v in self.args.set_attribute:
+					logging.debug('Setting attribute {}={}'.format(k, v))
+					ret = self._set_attribute(k, v, log_status=False)
+
+					if ret:
+						logging.debug('Unable to set attribute after flash command.')
+						logging.debug(ret)
+
 
 	def list_apps (self, verbose, quiet):
 		'''
@@ -391,47 +452,16 @@ class TockLoader:
 		'''
 		Change an attribute stored on the board.
 		'''
-		# Do some checking
-		if len(key.encode('utf-8')) > 8:
-			raise TockLoaderException('Key is too long. Must be 8 bytes or fewer.')
-		if len(value.encode('utf-8')) > 55:
-			raise TockLoaderException('Value is too long. Must be 55 bytes or fewer.')
 
 		# Enter bootloader mode to get things started
 		with self._start_communication_with_board():
 
-			if not self._bootloader_is_present():
-				raise TockLoaderException('No bootloader found! That means there is nowhere for attributes to go.')
+			# Use helper function to do all of the work.
+			ret = self._set_attribute(key, value)
 
-			# Create the buffer to write as the attribute
-			out = bytes([])
-			# Add key
-			out += key.encode('utf-8')
-			out += bytes([0] * (8-len(out)))
-			# Add length
-			out += bytes([len(value.encode('utf-8'))])
-			# Add value
-			out += value.encode('utf-8')
-
-			# Find if this attribute key already exists
-			open_index = -1
-			for index, attribute in enumerate(self.channel.get_all_attributes()):
-				if attribute:
-					if attribute['key'] == key:
-						logging.status('Found existing key at slot {}. Overwriting.'.format(index))
-						self.channel.set_attribute(index, out)
-						break
-				else:
-					# Save where we should put this attribute if it does not
-					# already exist.
-					if open_index == -1:
-						open_index = index
-			else:
-				if open_index == -1:
-					raise TockLoaderException('Error: No open space to save this attribute.')
-				else:
-					logging.status('Key not found. Writing new attribute to slot {}'.format(open_index))
-					self.channel.set_attribute(open_index, out)
+			if ret:
+				# Some error occurred!
+				raise TockLoaderException(ret)
 
 
 	def remove_attribute (self, key):
@@ -461,6 +491,20 @@ class TockLoader:
 				raise TockLoaderException('Error: Attribute does not exist.')
 
 
+	def set_start_address (self, address):
+		'''
+		Set the address that the bootloader jumps to to run kernel code.
+		'''
+
+		# Enter bootloader mode to get things started
+		with self._start_communication_with_board():
+
+			if not self._bootloader_is_present():
+				raise TockLoaderException('No bootloader found! That means there is nowhere for attributes to go.')
+
+			self.channel.set_start_address(address)
+
+
 	def info (self):
 		'''
 		Print all info about this board.
@@ -511,6 +555,15 @@ class TockLoader:
 			self._print_flash(address, flash)
 
 
+	def write_flash (self, address, length, value):
+		'''
+		Write a byte to some flash contents.
+		'''
+		with self._start_communication_with_board():
+			to_write = bytes([value] * length)
+			self.channel.flash_binary(address, to_write, pad=False)
+
+
 	def run_terminal (self):
 		'''
 		Create an interactive terminal session with the board.
@@ -604,8 +657,7 @@ class TockLoader:
 		length = 14
 		flag = self.channel.read_range(address, length)
 		flag_str = flag.decode('utf-8', 'ignore')
-		if self.args.debug:
-			logging.debug('Read from flags location: {}'.format(flag_str))
+		logging.debug('Read from flags location: {}'.format(flag_str))
 		return flag_str == 'TOCKBOOTLOADER'
 
 	def _update_board_specific_options (self):
@@ -634,9 +686,70 @@ class TockLoader:
 			# users do not have to pass them in every time.
 			if 'cmd_flags' in self.app_options:
 				for flag,setting in self.app_options['cmd_flags'].items():
-					logging.debug('Hardcoding command line argument "{}" to "{}" for board {}.'.format(flag, setting, board))
+					logging.info('Hardcoding command line argument "{}" to "{}" for board {}.'.format(flag, setting, board))
 					setattr(self.args, flag, setting)
 
+	############################################################################
+	## Helper Functions for Shared Code
+	############################################################################
+
+	def _set_attribute (self, key, value, log_status=True):
+		'''
+		Internal function for setting an attribute stored on the board.
+
+		Bootloader mode must be active.
+
+		Returns None if successful and an error string if not.
+		'''
+		# By default log status. However, that is not always appropriate, so
+		# if `log_status` is False then send that to debug output.
+		logging_fn = logging.status
+		if log_status == False:
+			logging_fn = logging.debug
+
+		# Do some checking
+		if len(key.encode('utf-8')) > 8:
+			return 'Key is too long. Must be 8 bytes or fewer.'
+		if len(value.encode('utf-8')) > 55:
+			return 'Value is too long. Must be 55 bytes or fewer.'
+
+		# Check for the bootloader, and importantly the attributes section.
+		if not self._bootloader_is_present():
+			return 'No bootloader found! That means there is nowhere for attributes to go.'
+
+		# Create the buffer to write as the attribute
+		out = bytes([])
+		# Add key
+		out += key.encode('utf-8')
+		out += bytes([0] * (8-len(out)))
+		# Add length
+		out += bytes([len(value.encode('utf-8'))])
+		# Add value
+		out += value.encode('utf-8')
+
+		# Find if this attribute key already exists
+		open_index = -1
+		for index, attribute in enumerate(self.channel.get_all_attributes()):
+			if attribute:
+				if attribute['key'] == key:
+					if attribute['value'] == value:
+						logging_fn('Found existing key,value at slot {}. Nothing to do.'.format(index))
+					else:
+						logging_fn('Found existing key at slot {}. Overwriting.'.format(index))
+						self.channel.set_attribute(index, out)
+					break
+			else:
+				# Save where we should put this attribute if it does not
+				# already exist.
+				if open_index == -1:
+					open_index = index
+		else:
+			if open_index == -1:
+				return 'No open space to save this attribute.'
+			else:
+				logging_fn('Key not found. Writing new attribute to slot {}'.format(open_index))
+				self.channel.set_attribute(open_index, out)
+
 	############################################################################
 	## Helper Functions for Manipulating Binaries and TBF
 	############################################################################
@@ -714,27 +827,34 @@ class TockLoader:
 			# flash. Also keep around the index of the app in original array.
 			slices = []
 			for i,app in enumerate(apps):
-				starting_addresses = app.get_fixed_addresses_flash()
+				apps_in_flash = app.get_fixed_addresses_flash_and_sizes()
 				app_slices = []
-				for sa in starting_addresses:
-					if sa < address:
+				for starting_address,size in apps_in_flash:
+					if starting_address < address:
 						# Can't use an app below the start of apps address.
 						continue
-					app_slices.append([sa, app.get_size(), i])
+					# HACK! Let's assume no board has more than 2 MB of flash.
+					if starting_address > (address + 0x200000):
+						logging.debug('Ignoring start address {:#x} as too large.'.format(starting_address))
+						continue
+					app_slices.append([starting_address, size, i])
 				slices.append(app_slices)
 
 			# See if we can find an ordering that works.
 			valid_order = brad_sort(slices)
 			if valid_order == None:
 				logging.error('Unable to find a valid sort order to flash apps.')
+				if self.args.debug:
+					for app in apps:
+						logging.debug('{}'.format(app.info(True)))
 				return
 
 			# Get sorted apps array.
-			logging.debug('Found sort order:')
+			logging.info('Found sort order:')
 			sorted_apps = []
 			for order in valid_order:
 				app = apps[order[2]]
-				logging.debug('  App "{}" at address {:#x}'.format(app.get_name(), order[0]))
+				logging.info('  App "{}" at address {:#x}'.format(app.get_name(), order[0]))
 				sorted_apps.append(app)
 			apps = sorted_apps
 
@@ -744,9 +864,9 @@ class TockLoader:
 			to_flash_apps = []
 			app_address = address
 			for app in apps:
-				# Ask the app where we can put it that is greater than or equal
-				# to the current address we are looking at.
-				next_loadable_address = app.get_next_loadable_address(app_address)
+				# Get a version of that app that we can put at a desirable address.
+				next_loadable_address = app.fix_at_next_loadable_address(app_address)
+
 				if next_loadable_address == app_address:
 					to_flash_apps.append(app)
 					app_address += app.get_size()
@@ -780,7 +900,7 @@ class TockLoader:
 				for app in to_flash_apps:
 					app_bundle += app.get_binary(app_address)
 					app_address += app.get_size()
-				logging.debug('Installing app bundle. Size: {} bytes.'.format(len(app_bundle)))
+				logging.info('Installing app bundle. Size: {} bytes.'.format(len(app_bundle)))
 				self.channel.flash_binary(address, app_bundle)
 			else:
 				# Flash only apps that have been modified. The only way an app
@@ -841,7 +961,7 @@ class TockLoader:
 			   (self.args.bundle_apps or
 			   	app.get_address() != app_address or
 			   	app.is_modified()):
-					logging.debug('Reading app {} binary from board.'.format(app))
+					logging.info('Reading app {} binary from board.'.format(app))
 					entire_app = self.channel.read_range(app.address, app.get_size())
 					in_flash_tbfh = TBFHeader(entire_app)
 					app.set_app_binary(entire_app[in_flash_tbfh.get_header_size():])
@@ -864,7 +984,7 @@ class TockLoader:
 			for app in apps:
 				app_bundle += app.get_binary(app_address)
 				app_address += app.get_size()
-			logging.debug('Installing app bundle. Size: {} bytes.'.format(len(app_bundle)))
+			logging.info('Installing app bundle. Size: {} bytes.'.format(len(app_bundle)))
 			self.channel.flash_binary(address, app_bundle)
 		else:
 			# Flash only apps that have been modified. The only way an app would
@@ -875,7 +995,7 @@ class TockLoader:
 				# the app is already installed.
 				optional_binary = app.get_binary(app_address)
 				if optional_binary:
-					logging.debug('Flashing app {} binary to board.'.format(app))
+					logging.info('Flashing app {} binary to board.'.format(app))
 					self.channel.flash_binary(app_address, optional_binary)
 				app_address = app_address + app.get_size()
 
@@ -934,32 +1054,37 @@ class TockLoader:
 		arch = self.channel.get_board_arch()
 
 		for tab in tabs:
-			if self.args.force or tab.is_compatible_with_board(self.channel.get_board_name()):
-				app = tab.extract_app(arch)
+			# Check if this app is specifically marked as compatible with
+			# certain boards, and if so, if the board being programmed is one of
+			# them.
+			if not self.args.force and not tab.is_compatible_with_board(self.channel.get_board_name()):
+				# App is marked for certain boards, and this is not one.
+				logging.info('App "{}" is not compatible with your board.'.format(tab.get_app_name()))
+				if self.args.debug:
+					logging.debug('Supported boards for app "{}":'.format(tab.get_app_name()))
+					for board in tab.get_compatible_boards():
+						logging.debug('- {}'.format(board))
+				continue
 
-				# Enforce the minimum app size here.
-				if app.get_size() < self.app_options['size_minimum']:
-					app.set_size(self.app_options['size_minimum'])
+			# Check if this app was compiled for the version of the Tock kernel
+			# currently on the board. If not, print a notice.
+			if not self.args.force and not tab.is_compatible_with_kernel_version(self.channel.get_kernel_version()):
+				# App needs a different kernel version than what is on the board.
+				logging.info('App "{}" requires kernel version "2", but tockloader determined your kernel version is "{}".'
+					.format(tab.get_app_name(), self.channel.get_kernel_version()))
+				continue
 
-				# Enforce other sizing constraints here.
-				if self.app_options['size_constraint'] == 'powers_of_two':
-					# Make sure the total app size is a power of two.
-					app_size = app.get_size()
-					if (app_size & (app_size - 1)) != 0:
-						# This is not a power of two, but should be.
-						count = 0
-						while app_size != 0:
-							app_size >>= 1
-							count += 1
-						app.set_size(1 << count)
-						if self.args.debug:
-							logging.debug('Rounding app up to ^2 size ({} bytes)'.format(1 << count))
-				elif self.app_options['size_constraint'] == 'none':
-					pass
-				else:
-					raise TockLoaderException('Unknown size constraint. This is a tockloader bug.')
+			# This app is good to install, continue the process.
 
-				apps.append(app)
+			app = tab.extract_app(arch)
+
+			# Enforce the minimum app size here.
+			app.set_minimum_size(self.app_options['size_minimum'])
+
+			# Enforce other sizing constraints here.
+			app.set_size_constraint(self.app_options['size_constraint'])
+
+			apps.append(app)
 
 		if len(apps) == 0:
 			raise TockLoaderException('No valid apps for this board were provided. Use --force to override.')
